from __future__ import annotations

import urllib.parse

from httpx import Client, Headers
from pydantic import BaseModel, Field

from ...image import (
    ImageAPI,
    ImageSearchResult,
    ImageSearchResultSet,
    ImageAPIContextManager,
)
from ..hookspec import hookimpl
from ..types import ImageAPIPlugin


PLUGIN_NAME = "unsplash"


class UnsplashBackendConfig(BaseModel):
    """
    Unsplash requires the usage of an ``access_key`` and ``secret_key``
    when using their API. We expose these settings here so users of the CLI
    tool can use it.
    """

    access_key: str = Field(description="Access key for the Unsplash API")

    secret_key: str = Field(description="Secret key for the Unsplash API")


CONFIG_FIELDS = {
    f"{PLUGIN_NAME}_config": (
        UnsplashBackendConfig,
        {"access_key": "", "secret_key": ""},
    )
}


class UnsplashImageAPI(ImageAPI):
    """
    Implementation of ImageAPI for use with the Unsplash API: https://unsplash.com/documentation
    """

    #: Base URL for the Unsplash API
    base_url = "https://api.unsplash.com/"

    #: Endpoint used for searching images
    search_endpoint = "/search/photos"

    def __init__(self, client_id: str, client: Client):
        """We use this initialization method to properly configure the ``httpx.Client`` object"""
        self._client_id = client_id
        self._headers = {"Authorization": f"Client-ID {client_id}"}
        self._client = client
        self._client.headers = Headers(self._headers)

    def search(self, query: str) -> ImageSearchResultSet:
        """
        Find images based on a ``search_term`` and return an ``ImageSearchResultSet``

        :raises HTTPError: Encountered during problems querying the API
        """
        search_url = urllib.parse.urljoin(self.base_url, self.search_endpoint)

        resp = self._client.get(search_url, params={"query": query})
        resp.raise_for_status()

        json_data = resp.json()

        search_results = tuple(
            ImageSearchResult(
                url=record.get("links", {}).get("download"),
                width=record.get("width"),
                height=record.get("height"),
            )
            for record in json_data.get("results", tuple())
        )

        return ImageSearchResultSet(search_results, json_data.get("total"))


class UnsplashImageAPIContextManager(ImageAPIContextManager):
    """
    Context manager that returns the ``UnsplashImageAPI`` we wish to use.
    This specific context manager handles setting up and tearing down the ``httpx.Client``
    connection that we use in this plugin.
    """

    def __enter__(self) -> UnsplashImageAPI:
        self.__client = Client()
        return UnsplashImageAPI(self._config.unsplash_config.access_key, self.__client)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__client.close()


@hookimpl
def image_api():
    """
    Registers our Unsplash image API backend
    """
    return ImageAPIPlugin(
        name=PLUGIN_NAME,
        image_api_context_manager=UnsplashImageAPIContextManager,
        config_fields=CONFIG_FIELDS,
    )



from pydantic import BaseSettings, Field

from ..constants import ENV_PREFIX


class BaseAppConfig(BaseSettings):
    """
    Holds all settings for the latz application. These are parsed from
    config files, command line arguments and environment variables.
    """

    backend: str = Field(
        default="unsplash",
        description="Image search backend to use for retrieving images.",
    )

    results_per_page: int = Field(
        default=10,
        description="Default number of images to display per page when searching.",
    )

    class Config:
        env_prefix = ENV_PREFIX



from __future__ import annotations

import json
import logging
from collections.abc import Sequence, Iterable
from pathlib import Path
from functools import reduce
from typing import NamedTuple, Any

from pydantic import ValidationError

from .models import BaseAppConfig
from .errors import format_validation_error, format_all_validation_errors

logger = logging.getLogger(__name__)


class ConfigError(Exception):
    pass


class ParsedConfigFile(NamedTuple):
    #: Path to the configuration file
    path: Path

    #: Data contained within the configuration file
    data: dict | None

    #: Parse errors encountered while parsing the configuration file
    error: str | None

    #: Parse errors encountered while parsing the configuration file
    model: BaseAppConfig | None


def merge_app_configs(
    app_configs: Iterable[BaseAppConfig], model_class: type[BaseAppConfig]
) -> BaseAppConfig:
    def reduce_func(
        config_one: BaseAppConfig, config_two: BaseAppConfig
    ) -> BaseAppConfig:
        """
        Provided two AppConfig objects, merge them into a single AppConfig object.
        Properties from ``config_two`` will override properties on ``config_one``
        """
        return model_class(**{**config_one.dict(), **config_two.dict()})

    return reduce(reduce_func, app_configs)


def parse_config_file_as_json(
    path: Path,
) -> ParsedConfigFile:
    """
    Given a path, try to parse it as JSON and ensure it is a dictionary
    and then return it. If it fails, we log a warning and return ``None``
    """
    try:
        with path.open() as fp:
            config_data_json = json.load(fp)
    except json.JSONDecodeError as exc:
        return ParsedConfigFile(
            data=None, path=path, error=f"Unable to parse {path}: {exc}", model=None
        )

    if not isinstance(config_data_json, dict):
        return ParsedConfigFile(
            data=None,
            path=path,
            error=f"Unable to parse {path}: JSON not correctly formatted",
            model=None,
        )

    return ParsedConfigFile(data=config_data_json, path=path, error=None, model=None)


def parse_app_config_model(
    parsed_config: ParsedConfigFile, model_class: type[BaseAppConfig]
) -> ParsedConfigFile:
    """
    Attempts to parse a dictionary as a ``AppConfig`` object. If successful,
    returns the ``AppConfig`` object and ``None`` as the error value. If not,
    returns ``None`` as the config and ``str`` as an error value.
    """
    try:
        if parsed_config.data:
            config_model = model_class(**parsed_config.data)
            return ParsedConfigFile(
                model=config_model, error=None, path=parsed_config.path, data=None
            )
        else:
            return ParsedConfigFile(
                model=None,
                path=parsed_config.path,
                error=parsed_config.error,
                data=None,
            )
    except ValidationError as exc:
        return ParsedConfigFile(
            model=None,
            error=format_validation_error(exc, parsed_config.path),
            data=parsed_config.data,
            path=parsed_config.path,
        )


def parse_config_files(paths: Sequence[Path]) -> tuple[ParsedConfigFile, ...] | None:
    """
    Given a list a ``paths`` to configuration files, returns those which
    can successfully be parsed. These are returned as ``ParsedConfigFile`` objects
    which contain the attributes "path", "data", and "error".
    """
    existing_paths = tuple(path for path in paths if path.is_file())

    if not existing_paths:
        return  # type: ignore

    # Parse JSON objects
    return tuple(parse_config_file_as_json(path) for path in existing_paths)


def get_app_config(
    paths: Sequence[Path], model_class: type[BaseAppConfig]
) -> BaseAppConfig:
    """
    Given a sequence of ``paths`` first attempts to parse these as JSON and then
    attempts to parse valid JSON objects as ``AppConfig`` objects.

    :raises ConfigError: Happens when any errors are encountered during config parsing
    """
    parsed_config_files = parse_config_files(paths)

    # No files were found 🤷‍; let's return a default config object
    if parsed_config_files is None:
        return model_class()

    parsed_config_files = tuple(
        parse_app_config_model(parsed, model_class) for parsed in parsed_config_files
    )

    # Gather errors
    errors = tuple(parsed.error for parsed in parsed_config_files if parsed.error)

    # Fail loudly if any errors
    if len(errors) > 0:
        raise ConfigError(format_all_validation_errors(errors))

    # Gather configs
    app_configs = tuple(parsed.model for parsed in parsed_config_files if parsed.model)

    # If there are no app_configs and no errors, provide a default
    if not app_configs:
        return model_class()

    # Merge all successfully parsed app_configs
    return merge_app_configs(app_configs, model_class)


def write_config_file(config_file_data: dict[str, Any], config_file: Path) -> None:
    """
    Attempts to write config file and returns the exception as a string if it failed.

    :raises ConfigError: Raised when we are not able to write our config file.
    """
    try:
        with config_file.open("w") as fp:
            json.dump(config_file_data, fp, indent=2)
    except OSError as exc:
        raise ConfigError(str(exc))



from pathlib import Path
from collections.abc import Sequence

from pydantic import ValidationError

#: Suggestion for where to find a documentation on configuration parameters
PARSE_ERROR_SUGGESTION = ""  # TODO: link to documentation

CONFIG_ERROR_PREFIX = "Unable to parse configuration file"


def format_validation_error(exc: ValidationError, path: Path) -> str:
    """
    Formats a ``pydantic.Validation`` error as a ``str``
    """
    error_str = []
    for err in exc.errors():
        loc = ",".join(err.get("loc", tuple()))  # type: ignore
        ctx = err.get("ctx", {})
        given = ctx.get("given")
        permitted = ctx.get("permitted")
        msg = err.get("msg")

        # Add strings
        error_str.append(f"\n  {loc}: \n")
        error_str.append(f"    {msg}\n")

        if given and permitted:
            error_str.append("  provided_value: ")
            error_str.append(f"'{given}'\n")

    return "".join(
        (f"{CONFIG_ERROR_PREFIX}: {path}\n", *error_str),
    )


def format_all_validation_errors(validation_errors: Sequence[str]) -> str:
    """
    Formats a sequence of errors as a single ``str``
    """
    error_str = "\n\n".join(validation_errors)
    # error_str += f"\n\n{PARSE_ERROR_SUGGESTION}"  #TODO: uncomment when no long blank string

    return error_str



from argparse import Namespace
from typing import cast

import rich_click as click
from pydantic import create_model, validator

from .commands import search_command, config_group
from .config import get_app_config, BaseAppConfig, ConfigError
from .constants import CONFIG_FILES
from .plugins.manager import get_plugin_manager

click.rich_click.USE_RICH_MARKUP = True
click.rich_click.USE_MARKDOWN_EMOJI = True


@click.group()
@click.pass_context
def cli(ctx):
    """
    Creates a click command group but also does the job of properly initializing
    our command line application.
    """
    ctx.ensure_object(Namespace)
    plugin_manager = get_plugin_manager()

    # We need to dynamically define our validators because we do not know all the of the
    # valid backends until runtime.
    validators = {
        "backend_validator": validator("backend", allow_reuse=True)(
            plugin_manager.get_backend_validator_func()
        )
    }

    # Dynamically create our new configuration object based on possible new fields
    # from our registered plugins.
    AppConfig = cast(
        type[BaseAppConfig],
        create_model(
            "AppConfig",
            **plugin_manager.image_api_config_fields,
            __validators__=validators,
            __base__=BaseAppConfig
        ),
    )

    # Creates the actual config object which parses all possible configuration sources
    # listed in ``CONFIG_FILES``.
    try:
        app_config = get_app_config(CONFIG_FILES, AppConfig)
    except ConfigError as exc:
        raise click.ClickException(str(exc))

    # Attach several properties to click's ``ctx`` object so that we have access to it
    # in our sub-commands.
    ctx.obj.config = app_config
    ctx.obj.image_api_context_manager = plugin_manager.get_image_api_context_manager(
        app_config
    )
    ctx.obj.config_class = AppConfig


cli.add_command(search_command)
cli.add_command(config_group)



from __future__ import annotations

import abc
from typing import Literal, NamedTuple


ImageTypes = Literal["png", "jpeg", "webp"]


class ImageSearchResult(NamedTuple):
    url: str | None
    width: int | None
    height: int | None


class ImageSearchResultSet(NamedTuple):
    results: tuple[ImageSearchResult, ...]
    total_number_results: int | None


class ImageAPIContextManager(abc.ABC):
    """
    Used to help plugin authors create compliant context managers.

    We use a class-based approach here to make sure that plugin authors appropriately
    implement the type of context manager the application needs (i.e. one that returns
    an implementation of ``ImageAPI``).
    """

    def __init__(self, config) -> None:
        """Used to initialize and configure the ``ImageAPI`` object"""
        self._config = config

    @abc.abstractmethod
    def __enter__(self) -> ImageAPI:
        """This must return an implementation of the below ``ImageAPI`` class"""

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Used to ensure the context manager protocol has been implemented"""


class ImageAPI(abc.ABC):
    """
    Abstract base class for the Image API
    """
    @abc.abstractmethod
    def search(self, query: str) -> ImageSearchResultSet:
        """
        Used to search for an image via a search term and then return a single result set
        """



from __future__ import annotations

import rich_click as click
from rich import print as rprint

from ...constants import CONFIG_FILE_CWD, CONFIG_FILE_HOME_DIR
from ...config import parse_config_file_as_json, write_config_file, ConfigError
from .validators import validate_and_parse_config_values


@click.group("config")
def group():
    """Various commands for showing and setting configuration values"""


@click.command("show")
@click.pass_context
def show_command(ctx):
    """Prints the current config"""
    rprint(ctx.obj.config.json(indent=2))


@click.command("set")
@click.argument("config_values", nargs=-1, callback=validate_and_parse_config_values)
@click.option(
    "-h",
    "--home",
    is_flag=True,
    help="Write to home directory config file instead of in current working directory",
)
def set_command(home, config_values):
    """
    Set configuration values.
    """
    config_file = CONFIG_FILE_HOME_DIR if home else CONFIG_FILE_CWD

    parsed_config = parse_config_file_as_json(config_file)

    if parsed_config.error is not None:
        raise click.ClickException(parsed_config.error)

    # Merge the new values and old values; new overwrites the old
    new_config_file_data = {**parsed_config.data, **config_values}

    try:
        write_config_file(new_config_file_data, config_file)
    except ConfigError as exc:
        raise click.ClickException(str(exc))


group.add_command(show_command)
group.add_command(set_command)



from collections.abc import Callable
from typing import cast, ContextManager, Any

import click
import httpx
from rich import print as rprint

from ..image import ImageAPI


@click.command("search")
@click.argument("query")
@click.pass_context
def command(ctx, query: str):
    """
    Command that retrieves an image based on a search term
    """
    image_api_context_manager = cast(
        Callable[[Any], ContextManager[ImageAPI]], ctx.obj.image_api_context_manager
    )

    with image_api_context_manager(ctx.obj.config) as api:
        try:
            result_set = api.search(query)
        except httpx.HTTPError as exc:
            raise click.ClickException(str(exc))

        for res in result_set.results:
            rprint(res)



from collections.abc import Iterable

import pluggy  # type: ignore

from ..constants import APP_NAME
from .types import ImageAPIPlugin

hookspec = pluggy.HookspecMarker(APP_NAME)
hookimpl = pluggy.HookimplMarker(APP_NAME)


class AppHookSpecs:
    """Holds all hookspecs for this application"""

    @hookspec
    def image_api(self) -> Iterable[ImageAPIPlugin]:
        """
        TODO: add example
        """
        return tuple()  # default implementation returns no registered backends



from typing import Literal
from urllib.parse import urljoin

from pydantic import BaseModel, Field

from ...image import (
    ImageAPI,
    ImageAPIContextManager,
    ImageSearchResult,
    ImageSearchResultSet,
)
from ..hookspec import hookimpl
from ..types import ImageAPIPlugin

PLUGIN_NAME = "dummy"


class DummyBackendConfig(BaseModel):
    """
    Unsplash requires the usage of an ``access_key`` and ``secret_key``
    when using their API. We expose these settings here so users of the CLI
    tool can use it.
    """

    placeholder_type: Literal["bear", "kitten"] = Field(
        default="kitten", description="The type of placeholder image links to use"
    )


CONFIG_FIELDS = {
    f"{PLUGIN_NAME}_config": (DummyBackendConfig, {"placeholder_type": "kitten"})
}


class DummyImageAPI(ImageAPI):
    """
    Used primarily to test and when we want to run the program without
    hitting any search endpoints.
    """

    def __init__(self, placeholder_type: str):
        self._placeholder_type = placeholder_type.lower()
        self._base_url = f"https://place{self._placeholder_type}.com"

    def search(self, query: str) -> ImageSearchResultSet:
        sizes = ((200, 300), (600, 500), (1000, 800))

        results = tuple(
            ImageSearchResult(
                url=urljoin(self._base_url, f"{width}/{height}"),
                width=width,
                height=height,
            )
            for width, height in sizes
        )

        return ImageSearchResultSet(results=results, total_number_results=len(results))


class DummyImageAPIContextManager(ImageAPIContextManager):
    """
    Dummy context manager for our DummyImageAPI
    """

    def __enter__(self) -> DummyImageAPI:
        return DummyImageAPI(self._config.dummy_config.placeholder_type)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


@hookimpl
def image_api():
    """
    Registers our Unsplash image API backend
    """
    return ImageAPIPlugin(
        name=PLUGIN_NAME,
        image_api_context_manager=DummyImageAPIContextManager,
        config_fields=CONFIG_FIELDS,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['latz',
 'latz.commands',
 'latz.commands.config',
 'latz.config',
 'latz.plugins',
 'latz.plugins.image']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.3.0,<10.0.0',
 'click>=8.1.3,<9.0.0',
 'httpx>=0.23.1,<0.24.0',
 'pluggy>=1.0.0,<2.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'rich-click>=1.6.0,<2.0.0',
 'rich>=12.6.0,<13.0.0']

entry_points = \
{'console_scripts': ['latz = latz.cli:cli']}

setup_kwargs = {
    'name': 'latz',
    'version': '0.1.2',
    'description': 'CLI Program for downloading images. Maybe by location too...',
    'long_description': '# latz\n\n[pluggy]: https://pluggy.readthedocs.io/en/stable/\n[click]: https://click.palletsprojects.com/\n[pydantic]: https://docs.pydantic.dev/\n[rich]: https://rich.readthedocs.io/\n[anaconda.org]: https://anaconda.org\n\nThis is a command line tool used for retrieving images from various image\nsearch backends (e.g. Unsplash, Google). This tool is primarily developed for educational purposes\nto show people how to develop plugin friendly Python applications. Furthermore,\nit is an example project that shows how to effectively pair a handful of\npopular Python libraries to write command line applications.\n\nTo facilitate our plugin architecture, the [pluggy][pluggy] library is used.\nOther libraries used include the following:\n\n- [click][click]: used for structuring the command line application 🖱 💻\n- [pydantic][pydantic]: used for handling configuration file validation 🗃\n- [rich][rich]: used for UX/UI elements and generally making the application more pretty 🌈\n\n### Why "latz"\n\n"latz" is short and easy to type! This is super important when writing CLI programs.\nI also thought about adding a geolocation search feature, so it is a reference\nto the word "latitude".\n\n## Quick Start\n\n### Installation\n\nlatz is available for install either on PyPI:\n\n```bash\n# Run from a new virtual environment\n$ pip install latz\n```\n\nor my own [anaconda.org][anaconda.org] channel:\n\n```bash\n$ conda create -n latz \'thath::latz\'\n```\n\nIf you are interested in tinkering around with the code yourself, you can also\nrun it locally:\n\n```bash\n$ git clone git@github.com:/travishathaway/latz.git\n$ cd latz\n# Create a virtual environment however you like..\n$ pip install -e .\n```\n\n### Configuring\n\nlatz comes initially configured with the "unsplash" image search backend. To use this,\nyou will need to create an Unsplash account and create a test application. After getting\nyour "access_key", you will need to configure it by adding it to your `.latz.json`\nconfig file. An example is show below:\n\n```json\n{\n  "backend": "unsplash",\n  "unsplash_config": {\n    "access_key": "your-access-key"\n  }\n}\n```\n\n_This file must be stored in your home directory or your current working directory._\n\nTo see other available image search backends, see [Available image search backends](#available-image-search-backends) below.\n\n### Usage\n\n_coming soon_ 😉\n\n### Available image search backends\n\n_coming soon_ 😉\n\n### How to extend and write your own image search backen\n\n_coming soon_ 😉\n',
    'author': 'Travis Hathaway',
    'author_email': 'travis.j.hathaway@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import NamedTuple, Any

from pydantic import BaseModel

from ..image import ImageAPIContextManager


class ImageAPIPlugin(NamedTuple):
    """
    Holds the metadata and class for the image search backend
    """

    name: str
    image_api_context_manager: type[ImageAPIContextManager]
    config_fields: dict[str, tuple[type[BaseModel], Any]]



import os
from pathlib import Path

#: Name of the application
APP_NAME = "latz"

#: Prefix used for environment variables
ENV_PREFIX = "LATZ_"

CONFIG_FILE_NAME = ".latz.json"

CONFIG_FILE_CWD = Path(os.getcwd()) / CONFIG_FILE_NAME

CONFIG_FILE_HOME_DIR = Path(os.path.expanduser("~")) / CONFIG_FILE_NAME

#: Config files to be loaded. Order will be respected, which means that
#: the config file on the bottom will override locations on the top.
CONFIG_FILES = (
    CONFIG_FILE_HOME_DIR,
    CONFIG_FILE_CWD,
)



from collections import Counter
from functools import reduce
from collections.abc import Callable

from click import ClickException
from pluggy import PluginManager  # type: ignore

from ..constants import APP_NAME
from ..config.models import BaseAppConfig
from ..image import ImageAPI
from .hookspec import AppHookSpecs
from .image import unsplash, dummy


class AppPluginManager(PluginManager):
    """
    Our own custom subclass of PluginManager. We do this to add several convenience methods
    we use elsewhere in our application.
    """

    def __init__(self, *args, **kwargs):
        # We use this to cache the value of
        self.__image_api_names = None

        super().__init__(*args, **kwargs)

    def reset_cache(self) -> None:
        """Resets all cached properties"""
        self.__image_api_names = None

    @property
    def image_api_names(self) -> tuple[str, ...]:
        """
        Get the names of available image search apis that are currently configured.
        We cache this once per run.

        If there are duplicate names registered, we raise an exception.

        :raises ClickException: Raised if duplicate values are found (same plugin
                                is registered multiple times)
        """
        if self.__image_api_names is not None:
            return self.__image_api_names

        self.__image_api_names = tuple(
            image_search_api.name for image_search_api in self.hook.image_api()
        )
        names_counter = Counter(self.__image_api_names)
        duplicates = tuple(value for value, count in names_counter.items() if count > 1)

        if len(duplicates) > 0:
            raise ClickException(
                "Duplicate values for the following 'image_search_api' found: "
                f"{', '.join(duplicates)}"
            )

        return self.__image_api_names

    @property
    def image_api_config_fields(self) -> dict:
        """
        Returns all the registered config fields for the image_search_api plugins.
        We perform a merge of all registered ``config_fields`` dictionaries that represent
        the new configuration fields that will be added.

        TODO: add a duplicate check just like for API names
        """
        return reduce(
            lambda dict_one, dict_two: {**dict_one, **dict_two},
            (
                image_search_api.config_fields
                for image_search_api in self.hook.image_api()
            ),
        )

    def get_image_api_context_manager(self, app_config: BaseAppConfig) -> ImageAPI:
        """
        Gets the currently configured api_context_manager. This is a context manager
        that we can use to give us an ``ImageAPI`` object.

        :param app_config: Current app config object
        :raises ClickException: Raised if we cannot find an api_context_manager object to return
        """
        image_api_context_manager = None

        for api in self.hook.image_api():
            if api.name == app_config.backend:
                image_api_context_manager = api.image_api_context_manager

        if image_api_context_manager is None:
            raise ClickException(
                "Backend has been improperly configured. Please choose from the available"
                f" backends: {', '.join(self.image_api_names)}"
            )

        return image_api_context_manager

    def get_backend_validator_func(self) -> Callable:
        """Returns the validator function that is used by Pydantic when parsing configuration"""

        def validate_backend(cls, value):
            if value not in self.image_api_names:
                valid_names = ", ".join(self.image_api_names)
                raise ValueError(
                    f"'{value}' is not valid choice for backend. "
                    f"Available choices: {valid_names}"
                )
            return value

        return validate_backend


def get_plugin_manager() -> AppPluginManager:
    """
    Plugin manager for the application. This function registers all plugin
    hooks, internal and external plugins.
    """
    # This is the plugin manager object we return that the application will use
    # to get information about register plugins.
    plugin_manager = AppPluginManager(APP_NAME)

    # Adds plugin hooks
    plugin_manager.add_hookspecs(AppHookSpecs)

    # Registers internal plugin hooks
    plugin_manager.register(unsplash)
    plugin_manager.register(dummy)

    # This is the magic that allows our application to discover other plugins
    # installed alongside it
    plugin_manager.load_setuptools_entrypoints(APP_NAME)

    return plugin_manager



from __future__ import annotations

import re
from typing import Any

import click
from pydantic import ValidationError


def validate_and_parse_config_values(ctx, _, values) -> dict[str, str]:
    """
    Validation function that parses and validates incoming configure parameter that
    look like this:
        param=value
        nested.param=value
    """
    pattern = re.compile(r"([a-z,A-Z,\.,_]+)=(.*)")

    current_config = ctx.obj.config.dict()
    checked_values = {}

    for value in values:
        match = pattern.match(value)
        if not match:
            raise click.BadParameter(
                f"'{value}' does not conform to the correct format. Please use the 'param=value'"
                " format."
            )
        parameter, value = match.groups()
        current_value = get_dotted_path_value(current_config, parameter)
        if not current_value:
            raise click.BadParameter(
                f"'{parameter}' is not a valid configuration parameter"
            )
        checked_values.update(get_nested_dict_from_path(parameter, value))

    try:
        ctx.obj.config_class(**{**current_config, **checked_values})
    except ValidationError as exc:
        raise click.ClickException(f"\n{str(exc)}")

    return checked_values


def get_dotted_path_value(nest: dict, dotted_path: str) -> Any:
    """
    When provides a ``dotted_path`` like "a.b.c" returns the nested value in ``nest``

    Example:
    >>> get_dotted_path_value({"a": {"b": {"c": 1}}}, "a.b.c.")
    1
    >>> get_dotted_path_value({"a": {"b": {"c": 1}}}, "a.b.c.d")  # We go as deep as we can
    1
    >>> get_dotted_path_value({"a": {"b": {"c": 1}}}, "d.c.b.a")

    """
    dotted_path_split = dotted_path.split(".")
    current_nest = nest

    for key_val in dotted_path_split:
        sub_val = current_nest.get(key_val)
        if not isinstance(sub_val, dict):
            return sub_val
        else:
            current_nest = sub_val


def get_nested_dict_from_path(path: str, value: Any) -> dict:
    """
    Used to transform our input config parameter "paths" (e.g. "nested.config.param")
    to an expanded dictionary.

    This only supports dictionaries that are structured like dict[str, dict[str, str]].

    Python's recursion limits prohibit deeply nested paths (> 999), but practically,
    this

    Examples:
        >>> get_nested_dict_from_path("nested.config.value", 123)
        {'nested': {'config': {'value': 123}}}
        >>> get_nested_dict_from_path("parameter", 123)
        {'parameter': 123}
        >>> get_nested_dict_from_path("", 123)
        {}
    """
    if not path:
        return {}

    def _inner(_parts: list[str], _value: Any) -> dict:
        if not _parts:
            return _value

        nested_dict = {}
        key = _parts.pop(0)
        nested_dict[key] = _inner(_parts, _value)

        return nested_dict

    parts = path.split(".")
    return _inner(parts, value)
